# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Benchmarking accuracy of qiskit to braket conversions

"""
import qiskit
import qiskit_braket_provider

import qbraid

from ..fixtures.qiskit.gates import get_qiskit_gates


def execute_test(conversion_function, input_circuit):
    """Execute conversion, test equality, and return 1 if it fails, 0 otherwise"""
    try:
        output_circuit = conversion_function(input_circuit)
        if not qbraid.interface.circuits_allclose(
            input_circuit, output_circuit, strict_gphase=False
        ):
            return 1
    except Exception:  # pylint: disable=broad-exception-caught
        return 1
    return 0


qiskit_gates = get_qiskit_gates(seed=0)
qbraid_failed, qiskit_failed = 0, 0

for gate_name, gate in qiskit_gates.items():
    if gate_name == "GlobalPhaseGate":
        continue

    qiskit_circuit = qiskit.QuantumCircuit(gate.num_qubits)
    qiskit_circuit.compose(gate, inplace=True)

    qiskit_failed += execute_test(
        qiskit_braket_provider.providers.adapter.convert_qiskit_to_braket_circuit, qiskit_circuit
    )
    qbraid_failed += execute_test(
        lambda circuit: qbraid.circuit_wrapper(circuit).transpile("braket"), qiskit_circuit
    )

total_tests = len(qiskit_gates)
qbraid_passed, qiskit_passed = total_tests - qbraid_failed, total_tests - qiskit_failed
qbraid_percentage, qiskit_percentage = round((qbraid_passed / total_tests) * 100, 2), round(
    (qiskit_passed / total_tests) * 100, 2
)

print("Qiskit to Braket standard gate conversion tests\n")
print(f"qbraid: {qbraid_passed}/{total_tests} ~= {qbraid_percentage}%")  # 58/58
print(f"qiskit-braket-provider: {qiskit_passed}/{total_tests} ~= {qiskit_percentage}%")  # 31/58



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for configuring provider credentials and authentication.

"""
from abc import ABC, abstractmethod
from typing import List

from qbraid._qdevice import QDEVICE, QDEVICE_TYPES


class QbraidDeviceNotFoundError(Exception):
    """Exception raised when no device could be found."""


class QuantumProvider(ABC):
    """
    This class is responsible for managing the interactions and
    authentications with various Quantum services.
    """

    @abstractmethod
    def save_config(self):
        """Save the current configuration."""

    @abstractmethod
    def get_devices(self):
        """Return a list of backends matching the specified filtering."""

    @abstractmethod
    def get_device(self, vendor_device_id: str):
        """Return quantum device corresponding to the specified device ID."""


class QbraidProvider(QuantumProvider):
    """
    This class is responsible for managing the interactions and
    authentications with the AWS and IBM Quantum services.

    Attributes:
        aws_access_key_id (str): AWS access key ID for authenticating with AWS services.
        aws_secret_access_key (str): AWS secret access key for authenticating with AWS services.
        qiskit_ibm_token (str): IBM Quantum token for authenticating with IBM Quantum services.
    """

    def __init__(self, aws_access_key_id=None, aws_secret_access_key=None, qiskit_ibm_token=None):
        """
        Initializes the QbraidProvider object with optional AWS and IBM Quantum credentials.

        Args:
            aws_access_key_id (str, optional): AWS access key ID. Defaults to None.
            aws_secret_access_key (str, optional): AWS secret access token. Defaults to None.
            qiskit_ibm_token (str, optional): IBM Quantum token. Defaults to None.
        """
        self._aws_provider = self._get_aws_provider(aws_access_key_id, aws_secret_access_key)
        self._ibm_provider = self._get_ibm_provider(qiskit_ibm_token)

    def save_config(self):
        raise NotImplementedError

    def _get_aws_provider(self, aws_access_key_id, aws_secret_access_key):
        if "braket.aws.aws_device.AwsDevice" not in QDEVICE_TYPES:
            return None

        from qbraid.providers.aws import BraketProvider  # pylint: disable=import-outside-toplevel

        try:
            return BraketProvider(aws_access_key_id, aws_secret_access_key)
        except Exception:  # pylint: disable=broad-exception-caught
            return None

    def _get_ibm_provider(self, qiskit_ibm_token):
        if "qiskit_ibm_provider.ibm_backend.IBMBackend" not in QDEVICE_TYPES:
            return None

        from qbraid.providers.ibm import QiskitProvider  # pylint: disable=import-outside-toplevel

        try:
            return QiskitProvider(qiskit_ibm_token)
        except Exception:  # pylint: disable=broad-exception-caught
            return None

    def _get_ibm_runtime(self, qiskit_ibm_token):
        if "qiskit_ibm_runtime.ibm_backend.IBMBackend" not in QDEVICE_TYPES:
            return None

        from qbraid.providers.ibm import QiskitRuntime  # pylint: disable=import-outside-toplevel

        try:
            return QiskitRuntime(qiskit_ibm_token)
        except Exception:  # pylint: disable=broad-exception-caught
            return None

    def get_devices(self) -> List[QDEVICE]:
        """Return a list of backends matching the specified filtering.

        Returns:
            list[QDEVICE]: a list of Backends that match the filtering
                criteria.
        """
        devices = []

        for provider in [self._aws_provider, self._ibm_provider]:
            if provider is not None:
                devices += provider.get_devices()

        return devices

    def get_device(self, vendor_device_id: str) -> QDEVICE:
        """Return quantum device corresponding to the specified device ID.

        Returns:
            QDEVICE: the quantum device corresponding to the given ID

        Raises:
            QbraidDeviceNotFoundError: if no device could be found
        """
        if vendor_device_id.startswith("ibm") or vendor_device_id.startswith("simulator"):
            if self._ibm_provider is not None:
                return self._ibm_provider.get_device(vendor_device_id)

        if vendor_device_id.startswith("arn:aws"):
            if self._aws_provider is not None:
                return self._aws_provider.get_device(vendor_device_id)

        raise QbraidDeviceNotFoundError(f"Device {vendor_device_id} not found.")



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module providing unified interface for interacting with
various quantum provider APIs.

.. currentmodule:: qbraid.providers

Data Types
------------

.. autosummary::
   :toctree: ../stubs/

   DeviceStatus
   DeviceType

Classes
--------

.. autosummary::
   :toctree: ../stubs/

   QuantumDevice
   QuantumJob
   JobStatus
   QbraidProvider
   QuantumJobResult

Exceptions
------------

.. autosummary::
   :toctree: ../stubs/

   JobError
   JobStateError
   ProgramValidationError
   QbraidDeviceNotFoundError
   QbraidRuntimeError

"""
from .device import QuantumDevice
from .enums import DeviceStatus, DeviceType, JobStatus
from .exceptions import JobError, JobStateError, ProgramValidationError, QbraidRuntimeError
from .job import QuantumJob
from .provider import QbraidDeviceNotFoundError, QbraidProvider
from .result import QuantumJobResult



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for inspecting and getting information about quantum programs

"""
from unittest.mock import Mock

import pytest

from qbraid.exceptions import PackageValueError, ProgramTypeError, QasmError
from qbraid.inspector import get_program_type, get_qasm_version

from ..fixtures import packages_bell

QASM_BELL_DATA = [
    (
        """
OPENQASM 2.0;
include "qelib1.inc";
qreg q[2];
h q[0];
cx q[0],q[1];
        """,
        "qasm2",
    ),
    (
        """
OPENQASM 3;
include "stdgates.inc";
qubit[2] q;
h q[0];
cx q[0], q[1];
        """,
        "qasm3",
    ),
    (
        """
OPENQASM 3.0;
bit[2] __bits__;
qubit[2] __qubits__;
h __qubits__[0];
cnot __qubits__[0], __qubits__[1];
__bits__[0] = measure __qubits__[0];
__bits__[1] = measure __qubits__[1];
        """,
        "qasm3",
    ),
]

QASM_ERROR_DATA = [
    """
OPENQASM 2.0;
include "qelib1.inc";
qreg q[2];
creg c[2];
h q[0;
cx q[0], q[1];
measure q[0] -> c[0];
measure q[1] -> c[1];

        """,
    """
OPENQASM 3.0;
include "stdgates.inc";
bit c[2];
qubit q[2];
h q[0]
cx q[0], q[1];
measure q -> c;
        """,
]


@pytest.mark.parametrize("qasm_str, expected_version", QASM_BELL_DATA)
def test_get_qasm_version(qasm_str, expected_version):
    """Test getting QASM version"""
    assert get_qasm_version(qasm_str) == expected_version


@pytest.mark.parametrize("qasm_str", QASM_ERROR_DATA)
def test_get_qasm_version_error(qasm_str):
    """Test getting QASM version"""
    with pytest.raises(QasmError):
        get_qasm_version(qasm_str)


@pytest.mark.parametrize("bell_circuit", packages_bell, indirect=True)
def test_get_program_type(bell_circuit):
    """Test that the correct package is returned for a given program."""
    circuit, expected_package = bell_circuit
    package_name = get_program_type(circuit)
    assert package_name == expected_package


def test_raise_error_unuspported_source_program():
    """Test that an error is raised if source program is not supported."""
    with pytest.raises(PackageValueError):
        get_program_type(Mock())


@pytest.mark.parametrize(
    "item",
    ["OPENQASM 2.0; bad operation", "OPENQASM 3.0; bad operation", "DECLARE ro BIT[1]", "circuit"],
)
def test_bad_source_openqasm_program(item):
    """Test raising ProgramTypeError converting invalid OpenQASM program string"""
    with pytest.raises(ProgramTypeError):
        get_program_type(item)


@pytest.mark.parametrize("item", [1, None])
def test_bad_source_program_type(item):
    """Test raising ProgramTypeError converting circuit of non-supported type"""
    with pytest.raises(ProgramTypeError):
        get_program_type(item)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.
"""
Module used for lazy loading of submodules.

"""
import importlib
import types


class LazyLoader(types.ModuleType):
    """Lazily loads a module upon attribute access.

    This class acts as a proxy for a module, loading it only when an attribute
    of the module is accessed for the first time.

    Attributes:
        module_name (str): The full qualified name of the module to be lazily loaded.
        parent_globals (dict): The globals of the parent module, where this loader is used.
        module (module, optional): The loaded module. Initially set to None.

    Args:
        module_name (str): The full qualified name of the module to be lazily loaded.
        parent_globals (dict): The globals of the parent module.
    """

    def __init__(self, module_name, parent_globals):
        super().__init__(module_name)
        self.module_name = module_name
        self.parent_globals = parent_globals
        self.module = None

    def _load(self):
        """Load the module and insert it into the parent's globals."""
        if not self.module:
            # Load the module and insert it into the parent's namespace
            self.module = importlib.import_module(self.module_name)
            self.parent_globals[self.__name__] = self.module

            # Update this object's dict for faster subsequent attribute access
            self.__dict__.update(self.module.__dict__)

        return self.module

    def __getattr__(self, item):
        return getattr(self._load(), item)

    def __dir__(self):
        return dir(self._load)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for performing QASM program checks before conversion

"""
from typing import TYPE_CHECKING

from openqasm3.parser import QASM3ParsingError, parse

from qbraid._qprogram import QPROGRAM_LIBS
from qbraid.exceptions import PackageValueError, ProgramTypeError, QasmError

if TYPE_CHECKING:
    import qbraid


def get_qasm_version(qasm_str: str) -> str:
    """Gets OpenQASM program version, either qasm2 or qasm3.

    Args:
        qasm_str: An OpenQASM program string

    Returns:
        QASM version from list :obj:`~qbraid.QPROGRAM_LIBS`

    Raises:
        :class:`~qbraid.QasmError`: If string does not represent a valid OpenQASAM program.

    """
    qasm_str = qasm_str.replace("opaque", "// opaque")

    try:
        program = parse(qasm_str)
        verion = int(float(program.version))
        return f"qasm{verion}"
    except QASM3ParsingError as err:
        raise QasmError("Failed to parse OpenQASM program.") from err


def get_program_type(program: "qbraid.QPROGRAM", require_supported: bool = True) -> str:
    """
    Get the type of a quantum program.

    Args:
        program (qbraid.QPROGRAM): The quantum program to get the type of.
        require_supported (bool): If True, raise an error if the program type is not supported.

    Returns:
        str: The type of the quantum program.
    """
    if isinstance(program, str):
        try:
            package = get_qasm_version(program)
        except QasmError as err:
            raise ProgramTypeError(
                "Input of type string must represent a valid OpenQASM program."
            ) from err

    else:
        try:
            program_module = program.__module__
            package = program_module.split(".")[0].lower()
        except AttributeError as err:
            raise ProgramTypeError(program) from err

    if package not in QPROGRAM_LIBS and require_supported:
        raise PackageValueError(package)

    return package



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining QiskitBackend Class

"""
import re
from typing import TYPE_CHECKING, Union  # pylint: disable=unused-import

from qiskit import transpile
from qiskit.transpiler import TranspilerError

from qbraid.providers.device import QuantumDevice
from qbraid.providers.enums import DeviceStatus, DeviceType

from .job import QiskitJob

if TYPE_CHECKING:
    import qiskit_ibm_provider
    import qiskit_ibm_runtime


class QiskitBackend(QuantumDevice):
    """Wrapper class for IBM Qiskit ``Backend`` objects."""

    def __init__(
        self, ibm_device: Union["qiskit_ibm_provider.IBMBackend", "qiskit_ibm_runtime.IBMBackend"]
    ):
        """Create a QiskitBackend."""

        super().__init__(ibm_device)
        self._vendor = "IBM"
        self._run_package = "qiskit"

    def _get_device_name(self) -> str:
        """Get the name of the device."""
        backend = self._device
        if hasattr(backend, "backend_name"):
            return backend.backend_name

        if hasattr(backend, "name"):
            return backend.name() if callable(backend.name) else backend.name

        match = re.search(r"<\w+\('([^']+)'\)>", str(backend))
        return match.group(1) if match else str(backend)

    def _populate_metadata(
        self, device: Union["qiskit_ibm_provider.IBMBackend", "qiskit_ibm_runtime.IBMBackend"]
    ) -> None:
        """Populate device metadata using IBMBackend object."""
        # pylint: disable=attribute-defined-outside-init
        device_name = self._get_device_name()
        self._id = device_name
        self._name = device_name
        self._provider = "IBM"
        if device_name.startswith("fake"):
            self._device_type = DeviceType("FAKE")
        else:
            self._device_type = DeviceType("SIMULATOR") if device.simulator else DeviceType("QPU")

        try:
            if self._device_type == DeviceType("FAKE"):
                self._num_qubits = device.configuration().n_qubits
            else:
                self._num_qubits = device.num_qubits
        except TranspilerError:
            if device.name == "simulator_stabilizer":
                self._num_qubits = 5000
            elif device.name == "simulator_extended_stabilizer":
                self._num_qubits = 63

            else:
                self._num_qubits = None

    def _transpile(self, run_input):
        if self._device_type.name != "FAKE" and self._device.local:
            # pylint: disable=import-outside-toplevel
            from qbraid.programs.qiskit import QiskitCircuit

            program = QiskitCircuit(run_input)
            program.remove_idle_qubits()
            run_input = program.program

        return transpile(run_input, backend=self._device)

    def _compile(self, run_input):
        return run_input

    def status(self):
        """Return the status of this Device.

        Returns:
            str: The status of this Device
        """
        if self._device_type == DeviceType("FAKE"):
            return DeviceStatus.ONLINE

        backend_status = self._device.status()
        if not backend_status.operational or backend_status.status_msg != "active":
            return DeviceStatus.OFFLINE
        return DeviceStatus.ONLINE

    def queue_depth(self) -> int:
        """Return the number of jobs in the queue for the ibm backend"""
        if self._device_type == DeviceType("FAKE"):
            return 0
        return self._device.status().pending_jobs

    def run(self, run_input, *args, **kwargs):
        """Runs circuit(s) on qiskit backend via :meth:`~qiskit.execute`

        Uses the :meth:`~qiskit.execute` method to create a :class:`~qiskit.providers.Job` object,
        applies a :class:`~qbraid.providers.ibm.QiskitJob`, and return the result.

        Args:
            run_input: A circuit object to run on the wrapped device.

        Keyword Args:
            shots (int): The number of times to run the task on the device. Default is 1024.

        Returns:
            qbraid.providers.ibm.QiskitJob: The job like object for the run.

        """
        backend = self._device
        qbraid_circuit = self.process_run_input(run_input)
        run_input = qbraid_circuit._program
        shots = kwargs.pop("shots", backend.options.get("shots"))
        memory = kwargs.pop("memory", True)  # Needed to get measurements
        qiskit_job = backend.run(run_input, shots=shots, memory=memory, **kwargs)
        try:
            tags_lst = qiskit_job.update_tags(kwargs.get("tags", []))
        except AttributeError:  # BasicAerJob does not have update_tags
            tags_lst = []
        tags = {tag: "*" for tag in tags_lst}
        qiskit_job_id = qiskit_job.job_id()
        qbraid_job_id = (
            self._init_job(qiskit_job_id, [qbraid_circuit], shots, tags)
            if self._device_type != DeviceType("FAKE")
            else "qbraid_test_id"
        )
        qbraid_job = QiskitJob(
            qbraid_job_id, vendor_job_id=qiskit_job_id, device=self, vendor_job_obj=qiskit_job
        )
        return qbraid_job

    def run_batch(self, run_input, **kwargs):
        """Runs circuit(s) on qiskit backend via :meth:`~qiskit.execute`

        Uses the :meth:`~qiskit.execute` method to create a :class:`~qiskit.providers.Job` object,
        applies a :class:`~qbraid.providers.ibm.QiskitJob`, and return the result.

        Args:
            run_input: A circuit object list to run on the wrapped device.

        Keyword Args:
            shots (int): The number of times to run the task on the device. Default is 1024.

        Returns:
            qbraid.providers.ibm.QiskitJob: The job like object for the run.

        """
        backend = self._device
        qbraid_circuit_batch = []
        run_input_batch = []
        for circuit in run_input:
            qbraid_circuit = self.process_run_input(circuit)
            run_input = qbraid_circuit._program
            run_input_batch.append(run_input)
            qbraid_circuit_batch.append(qbraid_circuit)

        shots = kwargs.pop("shots", backend.options.get("shots"))
        memory = kwargs.pop("memory", True)  # Needed to get measurements
        qiskit_job = backend.run(run_input_batch, shots=shots, memory=memory, **kwargs)
        try:
            tags_lst = qiskit_job.update_tags(kwargs.get("tags", []))
        except AttributeError:  # BasicAerJob does not have update_tags
            tags_lst = []
        tags = {tag: "*" for tag in tags_lst}
        qiskit_job_id = qiskit_job.job_id()

        # to change to batch
        qbraid_job_id = (
            self._init_job(qiskit_job_id, qbraid_circuit_batch, shots, tags)
            if self._device_type != DeviceType("FAKE")
            else "qbraid_test_id"
        )
        qbraid_job = QiskitJob(
            qbraid_job_id, vendor_job_id=qiskit_job_id, device=self, vendor_job_obj=qiskit_job
        )
        return qbraid_job



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for configuring provider credentials and authentication.

"""

import os
import re
from abc import abstractmethod
from typing import TYPE_CHECKING, List, Union

if TYPE_CHECKING:
    import qiskit_ibm_provider
    import qiskit_ibm_runtime


class QiskitRemoteService:
    """
    This class is responsible for managing the interactions and
    authentications with the IBM Quantum services.

    Attributes:
        qiskit_ibm_token (str): IBM Quantum token for authenticating with IBM Quantum services.
    """

    def __init__(self, qiskit_ibm_token=None, **kwargs):
        """
        Initializes the QbraidProvider object with optional AWS and IBM Quantum credentials.

        Args:
            qiskit_ibm_token (str, optional): IBM Quantum token. Defaults to None.
        """
        self.qiskit_ibm_token = qiskit_ibm_token or os.getenv("QISKIT_IBM_TOKEN")
        self._provider = self._get_ibm_provider(**kwargs)

    @abstractmethod
    def save_config(self):
        """Save the current configuration."""

    @abstractmethod
    def _get_ibm_provider(self, **kwargs):
        """Returns the IBM Quantum provider."""

    def get_devices(
        self, operational=True, **kwargs
    ) -> List[Union["qiskit_ibm_provider.IBMBackend", "qiskit_ibm_runtime.IBMBackend"]]:
        """Returns the IBM Quantum provider backends."""
        return self._provider.backends(operational=operational, **kwargs)

    def get_device(
        self, device_id: str
    ) -> Union["qiskit_ibm_provider.IBMBackend", "qiskit_ibm_runtime.IBMBackend"]:
        """Returns the IBM Quantum provider backends."""
        provider = self._get_ibm_provider()
        return provider.get_backend(device_id)

    @staticmethod
    def ibm_to_qbraid_id(name: str) -> str:
        """Converts IBM device name to qBraid device ID"""
        if name.startswith("ibm") or name.startswith("ibmq_"):
            return re.sub(r"^(ibm)(q)?_(.*)", r"\1_q_\3", name)
        return "ibm_q_" + name

    @abstractmethod
    def native_least_busy(
        self,
    ) -> Union["qiskit_ibm_provider.IBMBackend", "qiskit_ibm_runtime.IBMBackend"]:
        """Return the Backend object of the least busy qpu."""

    def ibm_least_busy_qpu(self) -> str:
        """Return the qBraid ID of the least busy IBMQ QPU."""
        backend = self.native_least_busy()
        ibm_id = backend.name  # QPU name of form `ibm_*` or `ibmq_*`
        _, name = ibm_id.split("_")
        return f"ibm_q_{name}"


class QiskitProvider(QiskitRemoteService):
    """Wrapper for qiskit_ibm_provider.IBMProvider class"""

    def save_config(self):
        """Save the current configuration."""
        raise NotImplementedError

    def _get_ibm_provider(self, **kwargs) -> "qiskit_ibm_provider.IBMProvider":
        """Returns the IBM Quantum provider."""
        from qiskit_ibm_provider import IBMProvider  # pylint: disable=import-outside-toplevel

        return IBMProvider(token=self.qiskit_ibm_token, **kwargs)

    def native_least_busy(self) -> "qiskit_ibm_provider.IBMBackend":
        """Return the qBraid ID of the least busy IBMQ QPU."""
        from qiskit_ibm_provider import least_busy  # pylint: disable=import-outside-toplevel

        backends = self.get_devices(simulator=False, operational=True)
        return least_busy(backends)


class QiskitRuntime(QiskitRemoteService):
    """Wrapper for qiskit_ibm_runtime.QiskitRuntimeService class."""

    def save_config(self):
        """Save the current configuration."""
        raise NotImplementedError

    def _get_ibm_provider(self, **kwargs) -> "qiskit_ibm_runtime.QiskitRuntimeService":
        """Returns the IBM Qiskt Runtime service."""
        # pylint: disable-next=import-outside-toplevel
        from qiskit_ibm_runtime import QiskitRuntimeService

        return QiskitRuntimeService(token=self.qiskit_ibm_token, **kwargs)

    def native_least_busy(self) -> "qiskit_ibm_runtime.IBMBackend":
        """Return the qBraid ID of the least busy IBMQ QPU."""
        return self._provider.least_busy(simulator=False, operational=True)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for conversions between Cirq Circuits and QASM strings

"""
from typing import Optional

import cirq
from cirq import ops
from cirq.contrib.qasm_import.exception import QasmException as CirqQasmException

import qbraid
from qbraid.exceptions import QasmError as QbraidQasmError
from qbraid.transpiler.conversions.cirq.cirq_gates import _map_zpow_and_unroll
from qbraid.transpiler.conversions.cirq.cirq_qasm_parser import QasmParser
from qbraid.transpiler.conversions.qasm_passes import flatten_qasm_program

QASMType = str


def _to_qasm_output(
    circuit: cirq.Circuit,
    header: Optional[str] = None,
    precision: int = 10,
    qubit_order: "cirq.QubitOrderOrList" = ops.QubitOrder.DEFAULT,
) -> "cirq.QasmOutput":
    """Returns a QASM object equivalent to the circuit.

    Args:
        header: A multi-line string that is placed in a comment at the top
            of the QASM. Defaults to a cirq version specifier.
        precision: Number of digits to use when representing numbers.
        qubit_order: Determines how qubits are ordered in the QASM
            register.
    """
    if header is None:
        header = f"Generated from qBraid v{qbraid._version.__version__}"
    qubits = ops.QubitOrder.as_qubit_order(qubit_order).order_for(circuit.all_qubits())
    return cirq.QasmOutput(
        operations=circuit.all_operations(),
        qubits=qubits,
        header=header,
        precision=precision,
        version="2.0",
    )


def cirq_to_qasm2(
    circuit: cirq.Circuit,
    header: Optional[str] = None,
    precision: int = 10,
    qubit_order: "cirq.QubitOrderOrList" = ops.QubitOrder.DEFAULT,
) -> QASMType:
    """Returns a QASM string representing the input Cirq circuit.

    Args:
        circuit: Cirq circuit to convert to a QASM string.

    Returns:
        QASMType: QASM string equivalent to the input Cirq circuit.
    """
    circuit = _map_zpow_and_unroll(circuit)
    return str(_to_qasm_output(circuit, header, precision, qubit_order))


def qasm2_to_cirq(qasm: QASMType) -> cirq.Circuit:
    """Returns a Cirq circuit equivalent to the input QASM string.

    Args:
        qasm: QASM string to convert to a Cirq circuit.

    Returns:
        Cirq circuit representation equivalent to the input QASM string.
    """
    try:
        qasm = flatten_qasm_program(qasm)
        return QasmParser().parse(qasm).circuit
    except CirqQasmException as err:
        raise QbraidQasmError from err



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining PytketCircuit Class

"""

from typing import TYPE_CHECKING, List, Optional, Union

import numpy as np
from pytket.circuit import Circuit, Command
from pytket.unit_id import Qubit

from qbraid.programs.abc_program import QuantumProgram

if TYPE_CHECKING:
    import pytket


class PytketCircuit(QuantumProgram):
    """Wrapper class for ``pytket.circuit.Circuit`` objects."""

    def __init__(self, program: "pytket.Circuit"):
        super().__init__(program)

    @property
    def program(self) -> Circuit:
        return self._program

    @program.setter
    def program(self, value: Circuit) -> None:
        if not isinstance(value, Circuit):
            raise ValueError("Program must be an instance of pytket.circuit.Circuit")
        self._program = value

    @property
    def qubits(self) -> List[Qubit]:
        """Return the qubits acted upon by the operations in this circuit"""
        return self.program.qubits

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits in the circuit."""
        return self.program.n_qubits

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return self.program.n_bits

    @property
    def depth(self) -> int:
        """Return the circuit depth (i.e., length of critical path)."""
        return self.program.depth()

    def _unitary(self) -> "np.ndarray":
        """Return the unitary of a pytket circuit."""
        return self.program.get_unitary()

    def remove_idle_qubits(self) -> None:
        """Checks whether the circuit uses contiguous qubits/indices,
        and if not, reduces dimension accordingly."""
        circuit = self.program.copy()
        circuit.remove_blank_wires()
        self._program = circuit

    def reverse_qubit_order(self) -> None:
        """Reverses the qubit ordering of a PyTKET circuit."""
        new_c = Circuit(self.num_qubits)
        for gate in self.program.get_commands():
            circuit_qubits = gate.qubits
            circuit_qubits = [(self.num_qubits - 1) - qubits.index[0] for qubits in circuit_qubits]
            gate_op = gate.op
            # devide parameter by pi, from radian to degree
            new_c.add_gate(
                gate_op.type,
                np.array(gate_op.params) / np.pi if gate_op.params else gate_op.params,
                circuit_qubits,
            )
        self._program = new_c

    @staticmethod
    def gate_to_matrix(
        gates: Optional[Union[List[Circuit], Command]], flat: bool = False
    ) -> np.ndarray:
        """Return the unitary of the Command"""
        gates = gates if (list == type(gates)) else [gates]
        a = list(map(max, [gate.qubits for gate in gates]))
        circuit = Circuit(max(a).index[0] + 1)
        for gate in gates:
            gate_op = gate.op
            circuit.add_gate(gate_op.type, gate_op.params, gate.qubits)
        if flat:
            circuit.remove_blank_wires()
        return circuit.get_unitary()



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining abstract QuantumJob Class

"""
import logging
from abc import ABC, abstractmethod
from enum import Enum
from time import sleep, time
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Union

from qbraid.api import QbraidSession

from .enums import JOB_FINAL, JobStatus
from .exceptions import JobError
from .status_maps import STATUS_MAP

if TYPE_CHECKING:
    import qbraid


class QuantumJob(ABC):
    """Abstract interface for job-like classes."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        job_id: str,
        vendor_job_id: Optional[str] = None,
        device: "Optional[qbraid.providers.QuantumDevice]" = None,
        vendor_job_obj: Optional[Any] = None,
        status: Optional[Union[str, JobStatus]] = None,
    ):
        self._vendor = None
        self._cache_metadata = None
        self._cache_status = self._map_status(status)
        self._job_id = job_id
        self._vendor_job_id = vendor_job_id
        self._device = device
        self._job = vendor_job_obj or self._get_job()
        self._status_map = STATUS_MAP[self.vendor]

    @property
    def id(self) -> str:  # pylint: disable=invalid-name
        """Return a unique id identifying the job."""
        return self._job_id

    @property
    def vendor(self) -> str:
        """Get job vendor."""
        if self._vendor is not None:
            return self._vendor

        try:
            vendor = self._cache_metadata["vendor"]
        except (KeyError, TypeError):
            vendor = self.device.vendor

        self._vendor = vendor.upper()
        return self._vendor

    @property
    def vendor_job_id(self) -> str:
        """Returns the ID assigned by the device vendor"""
        if self._vendor_job_id is None:
            self._cache_metadata = self._post_job_data()
            self._vendor_job_id = self._cache_metadata["vendorJobId"]
        return self._vendor_job_id

    @property
    def device(self) -> "qbraid.providers.QuantumDevice":
        """Returns the qbraid QuantumDevice object associated with this job."""
        if self._device is None:
            self._fetch_and_set_device()
        return self._device

    def _fetch_and_set_device(self) -> None:
        """Fetches device id from the server and sets the device object."""
        session = QbraidSession()
        job_lst = session.post(
            "/get-user-jobs", json={"qbraidJobId": self.id, "numResults": 1}
        ).json()

        if len(job_lst) == 0:
            job_lst = session.post("/get-user-jobs", json={"_id": self.id, "numResults": 1}).json()

            if len(job_lst) == 0:
                raise JobError(f"Could not find device associated with job {self.id}.")

        job_data = job_lst[0]

        try:
            import qbraid  # pylint: disable=import-outside-toplevel

            vendor_device_id = job_data["vendorDeviceId"]
            self._device = qbraid.device_wrapper(vendor_device_id)
        except Exception as err:  # pylint: disable=broad-except
            raise JobError(f"Could not find device associated with job {self.id}.") from err

    @staticmethod
    def _map_status(status: Optional[Union[str, JobStatus]] = None) -> JobStatus:
        """Returns `JobStatus` object mapped from raw status value. If no value
        provided or conversion from string fails, returns `JobStatus.UNKNOWN`."""
        if status is None:
            return JobStatus.UNKNOWN
        if isinstance(status, Enum):
            return status
        if isinstance(status, str):
            for e in JobStatus:
                status_enum = JobStatus(e.value)
                if status == status_enum.name or status == str(status_enum):
                    return status_enum
            raise ValueError(f"Status value '{status}' not recognized.")
        raise ValueError(f"Invalid status value type: {type(status)}")

    @staticmethod
    def status_final(status: Union[str, JobStatus]) -> bool:
        """Returns True if job is in final state. False otherwise."""
        if isinstance(status, str):
            if status in JOB_FINAL:
                return True
            for job_status in JOB_FINAL:
                if job_status.name == status:
                    return True
            return False
        raise TypeError(
            f"Expected status of type 'str' or 'JobStatus' \
            but instead got status of type {type(status)}."
        )

    def _status(self) -> Tuple[JobStatus, str]:
        vendor_status = self._get_status()
        try:
            return self._status_map[vendor_status], vendor_status
        except KeyError:
            logging.warning(
                "Expected %s job status matching one of %s but instead got '%s'.",
                self._device.vendor,
                str(list(self._status_map.keys())),
                vendor_status,
            )
            return JobStatus.UNKNOWN, vendor_status

    def status(self) -> JobStatus:
        """Return the status of the job / task , among the values of ``JobStatus``."""
        qbraid_status, vendor_status = self._status()
        if qbraid_status != self._cache_status:
            self._post_job_data(
                update={"status": vendor_status, "qbraidStatus": qbraid_status.name}
            )
        return qbraid_status

    def _post_job_data(self, update: Optional[dict] = None) -> dict:
        """Retreives job metadata and optionally updates document.

        Args:
            update: Dictionary containing fields to update in job document.

        Returns:
            The metadata associated with this job

        """
        session = QbraidSession()
        body = {"_id": self.id, "qbraidJobId": self.id}
        # Two status variables so we can track both qBraid and vendor status.
        if update is not None and "status" in update and "qbraidStatus" in update:
            body["status"] = update["status"]
            body["qbraidStatus"] = update["qbraidStatus"]
        metadata = session.put("/update-job", data=body).json()
        if "qbraidJobId" not in metadata:
            metadata["qbraidJobId"] = metadata.get("_id")
        return metadata

    def metadata(self) -> Dict[str, Any]:
        """Return the metadata regarding the job."""
        qbraid_status, vendor_status = self._status()
        if not self._cache_metadata or qbraid_status != self._cache_status:
            update = {"status": vendor_status, "qbraidStatus": qbraid_status.name}
            self._cache_metadata = self._post_job_data(update=update)
            self._cache_status = self._map_status(self._cache_metadata["qbraidStatus"])
        return self._cache_metadata

    def wait_for_final_state(self, timeout=None, poll_interval=5) -> None:
        """Poll the job status until it progresses to a final state.

        Args:
            timeout: Seconds to wait for the job. If ``None``, wait indefinitely.
            poll_interval: Seconds between queries.

        Raises:
            JobError: If the job does not reach a final state before the specified timeout.

        """
        start_time = time()
        status = self.status()
        while not self.status_final(status):
            elapsed_time = time() - start_time
            if timeout is not None and elapsed_time >= timeout:
                raise JobError(f"Timeout while waiting for job {self.id}.")
            sleep(poll_interval)
            status = self.status()

    @abstractmethod
    def _get_job(self):
        """Return the job like object that is being wrapped."""

    @abstractmethod
    def _get_status(self) -> str:
        """Returns job status casted as string."""

    @abstractmethod
    def result(self) -> "qbraid.providers.ResultWrapper":
        """Return the results of the job."""

    @abstractmethod
    def cancel(self) -> None:
        """Attempt to cancel the job."""

    def __repr__(self) -> str:
        """String representation of a QuantumJob object."""
        return f"<{self.__class__.__name__}(id:'{self.id}')>"



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for converting Braket circuits to/from OpenQASM 3

"""
import math
import re

from braket.circuits import Circuit
from braket.circuits.serialization import IRType
from braket.ir.openqasm import Program as OpenQasmProgram

from qbraid.exceptions import QasmError

QASMType = str


def _convert_pi_to_decimal(qasm_str: str) -> str:
    """Convert all instances of 'pi' in the QASM string to their decimal value."""

    pattern = r"(\d*\.?\d*\s*[*/+-]\s*)?pi(\s*[*/+-]\s*\d*\.?\d*)?"

    def replace_with_decimal(match):
        expr = match.group()
        try:
            value = eval(expr.replace("pi", str(math.pi)))  # pylint: disable=eval-used
        except SyntaxError:
            return expr
        return str(value)

    return re.sub(pattern, replace_with_decimal, qasm_str)


def qasm3_to_braket(qasm3_str: QASMType) -> Circuit:
    """Converts an OpenQASM 3.0 string to a ``braket.circuits.Circuit``.

    Args:
        qasm3_str: OpenQASM 3 string

    Returns:
        The Amazon Braket circuit equivalent to the input OpenQASM 3.0 string

    Raises:
        CircuitConversionError: If qasm to braket conversion fails

    """
    replacements = {
        "cx ": "cnot ",
        "sdg ": "si ",
        "tdg ": "ti ",
        "sx ": "v ",
        "sxdg ": "vi ",
        "p(": "phaseshift(",
        "cp(": "cphaseshift(",
    }

    def replace_commands(qasm, replacements):
        for old, new in replacements.items():
            qasm = qasm.replace(old, new)
        return qasm

    qasm3_str = qasm3_str.replace('include "stdgates.inc";', "")
    qasm3_str = replace_commands(qasm3_str, replacements)
    qasm3_str = _convert_pi_to_decimal(qasm3_str)

    try:
        program = OpenQasmProgram(source=qasm3_str)
        return Circuit.from_ir(source=program.source, inputs=program.inputs)
    except Exception as err:
        raise QasmError("Error converting qasm3 string to braket circuit") from err


def braket_to_qasm3(circuit: Circuit) -> QASMType:
    """Converts a ``braket.circuits.Circuit`` to an OpenQASM 3.0 string.

    Args:
        circuit: Amazon Braket quantum circuit

    Returns:
        The OpenQASM 3.0 string equivalent to the circuit

    Raises:
        CircuitConversionError: If braket to qasm conversion fails

    """
    try:
        return circuit.to_ir(IRType.OPENQASM).source
    except Exception as err:
        raise QasmError("Error converting braket circuit to qasm3 string") from err



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Helper functions for providers testing

"""
import braket.circuits
import cirq
import numpy as np
import qiskit

from qbraid.api import QbraidSession


def device_wrapper_inputs(vendor: str):
    """Returns list of tuples containing all device_wrapper inputs for given vendor."""
    session = QbraidSession()
    devices = session.get("/public/lab/get-devices", params={}).json()
    input_list = []
    skip_list = ["aws_ionq", "aws_ionq_forte1"]
    for document in devices:
        if document["vendor"] == vendor:
            qbraid_id = document["qbraid_id"]
            if qbraid_id not in skip_list:
                input_list.append(qbraid_id)
    return input_list


def braket_circuit():
    """Returns low-depth, one-qubit Braket circuit to be used for testing."""
    circuit = braket.circuits.Circuit()
    circuit.h(0)
    circuit.ry(0, np.pi / 2)
    return circuit


def cirq_circuit(meas=True):
    """Returns Low-depth, one-qubit Cirq circuit to be used for testing.
    If ``meas=True``, applies measurement operation to end of circuit."""
    q0 = cirq.GridQubit(0, 0)

    def basic_circuit():
        yield cirq.H(q0)
        yield cirq.Ry(rads=np.pi / 2)(q0)
        if meas:
            yield cirq.measure(q0, key="q0")

    circuit = cirq.Circuit()
    circuit.append(basic_circuit())
    return circuit


def qiskit_circuit(meas=True):
    """Returns Low-depth, one-qubit Qiskit circuit to be used for testing.
    If ``meas=True``, applies measurement operation to end of circuit."""
    circuit = qiskit.QuantumCircuit(1, 1) if meas else qiskit.QuantumCircuit(1)
    circuit.h(0)
    circuit.ry(np.pi / 2, 0)
    if meas:
        circuit.measure(0, 0)
    return circuit



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.
"""
Module for gathering system information.

"""

import os
import site
import sys

from qbraid._qdevice import QDEVICE_LIBS
from qbraid.exceptions import QbraidError


def get_site_packages_path() -> str:
    """Retrieves the site-packages path in the current Python environment."""
    site_packages_paths = site.getsitepackages()

    if len(site_packages_paths) == 1:
        return site_packages_paths[0]

    python_executable_path = sys.executable
    env_base_path = os.path.dirname(os.path.dirname(python_executable_path))

    for path in site_packages_paths:
        if env_base_path in path:
            return path

    raise QbraidError("Failed to find site-packages path.")


def _check_proxy(proxy_spec):
    """Checks if qBraid proxy is enabled for a given package and file."""
    package = proxy_spec[0]
    try:
        __import__(package)
    except ImportError:
        return False

    site_packages = get_site_packages_path()
    proxy_filepath = os.path.join(site_packages, *proxy_spec)

    try:
        with open(proxy_filepath, "r", encoding="utf-8") as file:
            for line in file:
                if "qbraid" in line:
                    return True
        return False
    except FileNotFoundError:
        return False
    except Exception:  # pylint: disable=broad-exception-caught
        return False


def qbraid_jobs_enabled(device_lib: str) -> bool:
    """Returns True if qBraid Quantum Jobs are enabled. Otherwise, returns False.

    See https://docs.qbraid.com/projects/lab/en/latest/lab/quantum_jobs.html
    """
    if device_lib not in QDEVICE_LIBS:
        return False

    if device_lib == "braket":
        proxy_spec = ("botocore", "httpsession.py")
    else:
        return False

    return _check_proxy(proxy_spec)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for converting Qiskit circuits to Cirq circuits.

"""
import numpy as np
import pytest
from qiskit import QuantumCircuit
from qiskit.circuit.random import random_circuit

from qbraid import circuit_wrapper
from qbraid.exceptions import QasmError
from qbraid.interface import circuits_allclose
from qbraid.transpiler.conversions.cirq.conversions_qasm import qasm2_to_cirq
from qbraid.transpiler.conversions.qiskit.conversions_qasm import qiskit_to_qasm2
from qbraid.transpiler.converter import convert_to_package


def test_bell_state_from_qiskit():
    """Tests qiskit.QuantumCircuit --> cirq.Circuit
    with a Bell state circuit.
    """
    qiskit_circuit = QuantumCircuit(2)
    qiskit_circuit.h(0)
    qiskit_circuit.cx(0, 1)
    cirq_circuit = convert_to_package(qiskit_circuit, "cirq")
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


def test_common_gates_from_qiskit():
    """Tests converting standard gates from Qiskit to Cirq."""
    qiskit_circuit = QuantumCircuit(4)
    qiskit_circuit.h([0, 1, 2, 3])
    qiskit_circuit.x([0, 1])
    qiskit_circuit.y(2)
    qiskit_circuit.z(3)
    qiskit_circuit.s(0)
    qiskit_circuit.sdg(1)
    qiskit_circuit.t(2)
    qiskit_circuit.tdg(3)
    qiskit_circuit.rx(np.pi / 4, 0)
    qiskit_circuit.ry(np.pi / 2, 1)
    qiskit_circuit.rz(3 * np.pi / 4, 2)
    qiskit_circuit.p(np.pi / 8, 3)
    qiskit_circuit.sx(0)
    qiskit_circuit.sxdg(1)
    qiskit_circuit.iswap(2, 3)
    qiskit_circuit.swap([0, 1], [2, 3])
    qiskit_circuit.cx(0, 1)
    qiskit_circuit.cp(np.pi / 4, 2, 3)

    cirq_circuit = convert_to_package(qiskit_circuit, "cirq")
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("qubits", ([0, 1], [1, 0]))
def test_crz_gate_from_qiskit(qubits):
    """Tests converting controlled Rz gate from Qiskit to Cirq."""
    qiskit_circuit = QuantumCircuit(2)
    qiskit_circuit.crz(np.pi / 4, *qubits)
    cirq_circuit = convert_to_package(qiskit_circuit, "cirq")
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("qubits", ([0, 1], [1, 0]))
@pytest.mark.parametrize("theta", (0, 2 * np.pi, np.pi / 2, np.pi / 4))
def test_rzz_gate_from_qiskit(qubits, theta):
    """Tests converting Rzz gate from Qiskit to Cirq."""
    qiskit_circuit = QuantumCircuit(2)
    qiskit_circuit.rzz(theta, *qubits)
    cirq_circuit = convert_to_package(qiskit_circuit, "cirq")
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


def test_iswap_gate_from_qiskit():
    """Tests converting iSwap gate from Qiskit to Cirq."""
    qiskit_circuit = QuantumCircuit(2)
    qiskit_circuit.h([0, 1])
    qiskit_circuit.iswap(0, 1)
    cirq_circuit = convert_to_package(qiskit_circuit, "cirq")
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


def test_qiskit_roundtrip():
    """Test converting qiskit gates that previously failed qiskit roundtrip test."""
    qiskit_circuit = QuantumCircuit(3)
    qiskit_circuit.ccz(0, 1, 2)
    qiskit_circuit.ecr(1, 2)
    qiskit_circuit.cs(2, 0)
    cirq_circuit = convert_to_package(qiskit_circuit, "cirq")
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=False)


def test_qiskit_roundtrip_noncontig():
    """Test converting gates that previously failed qiskit roundtrip test
    with non-contiguous qubits."""
    qiskit_circuit = QuantumCircuit(4)
    qiskit_circuit.ccz(0, 1, 2)
    qiskit_circuit.ecr(1, 2)
    qiskit_circuit.cs(2, 0)
    cirq_circuit = convert_to_package(qiskit_circuit, "cirq")
    qprogram = circuit_wrapper(cirq_circuit)
    qprogram.remove_idle_qubits()
    qiskit_contig = qprogram.program
    assert circuits_allclose(qiskit_contig, cirq_circuit, strict_gphase=False)


def test_100_random_qiskit():
    """Test converting 100 random qiskit circuits to cirq."""
    for _ in range(100):
        qiskit_circuit = random_circuit(4, 1)
        cirq_circuit = convert_to_package(qiskit_circuit, "cirq")
        assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=False)


def test_raise_error():
    """Test raising error for unsupported gates."""
    with pytest.raises(QasmError):
        qiskit_circuit = QuantumCircuit(1)
        qiskit_circuit.delay(300, 0)
        qasm2 = qiskit_to_qasm2(qiskit_circuit)
        _ = qasm2_to_cirq(qasm2)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for emitting and disabling warnings at top level.

"""
import warnings
from importlib.metadata import PackageNotFoundError, version

import requests
import urllib3

from .exceptions import QbraidError


def _get_latest_version(package: str) -> str:
    """Retrieves the latest version of a package from PyPI."""
    url = f"https://pypi.org/pypi/{package}/json"
    try:
        response = requests.get(url, timeout=5)
        response.raise_for_status()
        data = response.json()
        return data["info"]["version"]
    except requests.RequestException as err:
        raise QbraidError(f"Failed to retrieve latest {package} version.") from err


def _get_local_version(package: str) -> str:
    """Retrieves the local version of a package."""
    try:
        return version(package)
    except PackageNotFoundError as err:
        raise QbraidError(f"{package} is not installed in the current environment.") from err


def _warn_new_version(local: str, latest: str) -> bool:
    """Returns True if you should warn user about updated package version,
    False otherwise."""
    installed_major, installed_minor = map(int, local.split(".")[:2])
    latest_major, latest_minor = map(int, latest.split(".")[:2])

    return (installed_major, installed_minor) < (latest_major, latest_minor)


def _check_version():
    """Emits UserWarning if updated package version exists in qBraid API
    compared to local copy."""

    # pylint: disable=import-outside-toplevel
    try:
        latest_version = _get_latest_version("qbraid")
        local_version = _get_local_version("qbraid")

        if _warn_new_version(local_version, latest_version):
            warnings.warn(
                f"You are using qbraid version {local_version}; however, version {latest_version} "
                "is available. To avoid compatibility issues, consider upgrading. ",
                UserWarning,
            )
    except QbraidError:
        pass


# coverage: ignore
warnings.filterwarnings("ignore", category=SyntaxWarning)
warnings.filterwarnings("ignore", category=UserWarning, message="Setuptools is replacing distutils")
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=PendingDeprecationWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning, module="numpy")
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

_check_version()



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Test circuit equality helper functions

"""

import numpy as np
import pytest

from qbraid.interface.circuit_equality import assert_allclose_up_to_global_phase, match_global_phase


def test_match_global_phase_basic():
    """Test matching global phase of two arrays"""
    a = np.array([1 + 1j, 2 + 2j])
    b = np.array([1 - 1j, 2 - 2j])
    a_prime, b_prime = match_global_phase(a, b)
    assert np.allclose(a_prime, b_prime)


def test_match_global_phase_shape_mismatch():
    """Test matching global phase of two arrays with shape mismatch"""
    a = np.array([1 + 1j, 2 + 2j])
    b = np.array([[1 - 1j, 2 - 2j]])
    a_prime, b_prime = match_global_phase(a, b)
    assert np.array_equal(a, a_prime) and np.array_equal(b, b_prime)


def test_assert_allclose_up_to_global_phase_basic():
    """Test assert_allclose_up_to_global_phase with two arrays"""
    a = np.array([1 + 1j, 2 + 2j])
    b = np.array([1 - 1j, 2 - 2j])
    assert_allclose_up_to_global_phase(a, b, atol=1e-10)


def test_assert_allclose_up_to_global_phase_fail():
    """Test assert_allclose_up_to_global_phase with two arrays that are not close"""
    a = np.array([1 + 1j, 2 + 2j])
    b = np.array([10 + 10j, 20 + 20j])
    with pytest.raises(AssertionError):
        assert_allclose_up_to_global_phase(a, b, atol=1e-10)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for quantum jobs functions and data types

"""
import pytest

from qbraid import job_wrapper
from qbraid.exceptions import QbraidError
from qbraid.providers.enums import JobStatus
from qbraid.providers.job import QuantumJob

status_data = [
    (JobStatus.INITIALIZING, False),
    (JobStatus.QUEUED, False),
    (JobStatus.RUNNING, False),
    (JobStatus.CANCELLING, False),
    (JobStatus.CANCELLED, True),
    (JobStatus.COMPLETED, True),
    (JobStatus.FAILED, True),
    (JobStatus.UNKNOWN, False),
]


@pytest.mark.parametrize("status_tuple", status_data)
def test_is_status_final(status_tuple):
    """Test identifying job status objects that are in final state"""
    status_obj, final_expected = status_tuple
    status_str = status_obj.name
    final_obj = QuantumJob.status_final(status_obj)
    final_str = QuantumJob.status_final(status_str)
    assert final_obj == final_expected
    assert final_str == final_expected


@pytest.mark.parametrize("status", [0, {}, []])
def test_is_status_final_error(status):
    """Test raiseing exception while checking if job status is final"""
    with pytest.raises(TypeError):
        QuantumJob.status_final(status)


@pytest.mark.parametrize("status_tuple", status_data)
def test_status_from_raw(status_tuple):
    """Test converting str status representation to JobStatus object."""
    status_obj, _ = status_tuple
    status_str = status_obj.name
    status_obj_test = QuantumJob._map_status(status_str)
    assert status_obj == status_obj_test


@pytest.mark.parametrize("status", ["INITIIZING", "QUDEUED", "CANCELLLED"])
def test_status_from_raw_error(status):
    """Test raising exception while converting invalid str status representation."""
    with pytest.raises(ValueError):
        QuantumJob._map_status(status)


@pytest.mark.parametrize(
    "status_tuple",
    [
        (JobStatus.QUEUED, JobStatus.QUEUED),
        ("VALIDATING", JobStatus.VALIDATING),
        (None, JobStatus.UNKNOWN),
    ],
)
def test_map_status(status_tuple):
    """Test setting initial job status."""
    status_input, status_expected = status_tuple
    assert QuantumJob._map_status(status_input) == status_expected


def test_map_status_raises():
    """Test raising exception for bad status type."""
    with pytest.raises(ValueError):
        QuantumJob._map_status(0)


def test_job_wrapper_id_error():
    """Test raising job wrapper error due to invalid job ID."""
    with pytest.raises(QbraidError):
        job_wrapper("Id123")



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for QiskitBackend class.

"""
import os
import time
from typing import Union

import pytest
from qiskit.providers import Backend
from qiskit.providers.basicaer.basicaerjob import BasicAerJob
from qiskit.providers.fake_provider import FakeManila, FakeProviderFactory
from qiskit_aer.jobs.aerjob import AerJob
from qiskit_ibm_provider import IBMBackend, IBMJob

from qbraid import device_wrapper
from qbraid.providers import QuantumJob
from qbraid.providers.exceptions import JobStateError
from qbraid.providers.ibm import QiskitBackend, QiskitJob, QiskitProvider

from .fixtures import cirq_circuit, device_wrapper_inputs, qiskit_circuit

# Skip tests if IBM account auth/creds not configured
skip_remote_tests: bool = os.getenv("QBRAID_RUN_REMOTE_TESTS") is None
REASON = "QBRAID_RUN_REMOTE_TESTS not set (requires configuration of IBM storage)"


@pytest.fixture
def providers():
    """Return tuple of IBM and fake providers."""
    ibmq_token = os.getenv("QISKIT_IBM_TOKEN", None)
    qbraid_provider = QiskitProvider(qiskit_ibm_token=ibmq_token)
    ibm_provider = qbraid_provider._provider
    fake_provider = FakeProviderFactory().get_provider()
    return ibm_provider, fake_provider


def ibm_devices():
    """Get list of wrapped ibm backends for testing."""
    provider = QiskitProvider()
    backends = provider.get_devices(
        filters=lambda b: b.status().status_msg == "active", operational=True
    )
    qbraid_device_names = device_wrapper_inputs("IBM")
    ibm_device_names = [provider.ibm_to_qbraid_id(backend.name) for backend in backends]
    return [dev for dev in ibm_device_names if dev in qbraid_device_names]


def fake_ibm_devices():
    """Get list of fake wrapped ibm backends for testing"""
    fake_provider = FakeProviderFactory().get_provider()
    backends = fake_provider.backends()
    return [QiskitBackend(backend) for backend in backends if backend.configuration().n_qubits < 24]


inputs_qiskit_dw = [] if skip_remote_tests else ibm_devices()
circuits_qiskit_run = [cirq_circuit(), qiskit_circuit()]


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
@pytest.mark.parametrize("device_id", inputs_qiskit_dw)
def test_device_wrapper_ibm_from_api(device_id):
    """Test creating device wrapper from Qiskit device ID."""
    qbraid_device = device_wrapper(device_id)
    vendor_device = qbraid_device._device
    assert isinstance(qbraid_device, QiskitBackend)
    assert isinstance(vendor_device, IBMBackend)


def test_wrap_fake_provider():
    """Test wrapping fake Qiskit provider."""
    backend = FakeManila()
    backend.simulator = True
    qbraid_device = QiskitBackend(backend)
    vendor_device = qbraid_device._device
    assert isinstance(qbraid_device, QiskitBackend)
    assert isinstance(vendor_device, Backend)


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
def test_queue_depth():
    """Test getting number of pending jobs for QiskitBackend."""
    ibm_device = device_wrapper("ibm_q_qasm_simulator")
    assert isinstance(ibm_device.queue_depth(), int)


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
@pytest.mark.parametrize("circuit", circuits_qiskit_run)
def test_run_qiskit_device_wrapper(circuit):
    """Test run method from wrapped Qiskit backends"""
    qbraid_device = device_wrapper("ibm_q_qasm_simulator")
    qbraid_job = qbraid_device.run(circuit, shots=10)
    vendor_job = qbraid_job._job
    assert isinstance(qbraid_job, QiskitJob)
    assert isinstance(vendor_job, IBMJob)


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
def test_run_batch_qiskit_device_wrapper():
    """Test run_batch method from wrapped Qiskit backends"""
    qbraid_device = device_wrapper("ibm_q_qasm_simulator")
    qbraid_job = qbraid_device.run_batch(circuits_qiskit_run, shots=10)
    vendor_job = qbraid_job._job
    assert isinstance(qbraid_job, QiskitJob)
    assert isinstance(vendor_job, IBMJob)


@pytest.mark.parametrize("qbraid_device", fake_ibm_devices())
@pytest.mark.parametrize("circuit", circuits_qiskit_run)
def test_run_fake_qiskit_device_wrapper(qbraid_device, circuit):
    """Test run method from wrapped fake Qiskit backends"""
    qbraid_job = qbraid_device.run(circuit, shots=10)
    vendor_job = qbraid_job._job
    assert isinstance(qbraid_job, QiskitJob)
    assert isinstance(vendor_job, Union[BasicAerJob, AerJob])


@pytest.mark.parametrize("qbraid_device", fake_ibm_devices())
def test_run_fake_batch_qiskit_device_wrapper(qbraid_device):
    """Test run method from wrapped fake Qiskit backends"""
    qbraid_job = qbraid_device.run_batch(circuits_qiskit_run, shots=10)
    vendor_job = qbraid_job._job
    assert isinstance(qbraid_job, QiskitJob)
    assert isinstance(vendor_job, Union[BasicAerJob, AerJob])


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
def test_cancel_completed_batch_error():
    """Test that cancelling a batch job that has already reached its
    final state raises an error."""

    # Initialize your quantum device
    qbraid_device = device_wrapper("ibm_q_simulator_statevector")
    qbraid_job = qbraid_device.run_batch(circuits_qiskit_run, shots=10)

    # Initialize your timer
    timeout = 60  # Total time to wait for job to complete
    check_every = 2  # Check job status every 2 seconds
    elapsed_time = 0

    # Wait for job to complete, but not more than the timeout
    while elapsed_time < timeout:
        # Check if job has reached its final state
        status = qbraid_job.status()
        if QuantumJob.status_final(status):
            break

        # If not, sleep and then check again
        time.sleep(check_every)
        elapsed_time += check_every

    # If job hasn't finished even after waiting for the timeout period, cancel it
    if elapsed_time >= timeout:
        try:
            qbraid_job.cancel()
        except JobStateError:
            pass

    # Ensure that cancelling a finished job raises an error
    with pytest.raises(JobStateError):
        qbraid_job.cancel()


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
@pytest.mark.parametrize("backend_name", ["ibmq_qasm_simulator", "fake_melbourne", "fake_almaden"])
def test_get_device_name_fake(providers, backend_name):  # pylint: disable=redefined-outer-name
    """Test edge cases for getting device name, e.g. .name, .name(), .backend_name"""
    provider, fake_provider = providers

    if backend_name.startswith("ibmq_"):
        backend = provider.get_backend(backend_name)
    else:
        backend = fake_provider.get_backend(backend_name)

    qbraid_device = QiskitBackend(backend)

    assert qbraid_device._get_device_name() == backend_name



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for converting pytket circuits to Cirq circuits.

"""
import numpy as np
import pytest
from pytket.circuit import Circuit as TKCircuit

from qbraid.interface import circuits_allclose, random_circuit
from qbraid.transpiler import convert_to_package
from qbraid.transpiler.exceptions import CircuitConversionError


def test_bell_state_from_qiskit():
    """Tests pytket.circuit.Circuit --> cirq.Circuit
    with a Bell state circuit."""
    pytket_circuit = TKCircuit(2)
    pytket_circuit.H(0)
    pytket_circuit.CX(0, 1)
    cirq_circuit = convert_to_package(pytket_circuit, "cirq")
    assert circuits_allclose(pytket_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("qubits", ([0, 1], [1, 0]))
def test_crz_gate_from_pytket(qubits):
    """Test converting controlled Rz gate from pytket to cirq."""
    pytket_circuit = TKCircuit(2)
    pytket_circuit.CRz(np.pi / 4, *qubits)
    cirq_circuit = convert_to_package(pytket_circuit, "cirq")
    assert circuits_allclose(pytket_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("qubits", ([0, 1], [1, 0]))
@pytest.mark.parametrize("theta", (0, 2 * np.pi, np.pi / 2, np.pi / 4))
def test_rzz_gate_from_pytket(qubits, theta):
    """Test converting Rzz gate from pytket to cirq."""
    pytket_circuit = TKCircuit(2)
    pytket_circuit.ZZPhase(theta, *qubits)
    cirq_circuit = convert_to_package(pytket_circuit, "cirq")
    assert circuits_allclose(pytket_circuit, cirq_circuit, strict_gphase=True)


def test_100_random_pytket():
    """Test converting 100 random pytket circuits to cirq."""
    for _ in range(100):
        pytket_circuit = random_circuit("pytket", 4, 1)
        cirq_circuit = convert_to_package(pytket_circuit, "cirq")
        assert circuits_allclose(pytket_circuit, cirq_circuit, strict_gphase=False)


def test_raise_error():
    """Test raising an error when converting an unsupported gate."""
    with pytest.raises(CircuitConversionError):
        pytket_circuit = TKCircuit(2)
        pytket_circuit.ISWAPMax(0, 1)
        convert_to_package(pytket_circuit, "cirq")



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining PyQuilProgram Class

"""

from typing import List

import numpy as np
import pyquil
from pyquil.simulation.tools import program_unitary

from qbraid.programs.abc_program import QuantumProgram


class PyQuilProgram(QuantumProgram):
    """Wrapper class for ``pyQuil.Program`` objects."""

    def __init__(self, program: "pyquil.Program"):
        super().__init__(program)

    @property
    def program(self) -> pyquil.Program:
        return self._program

    @program.setter
    def program(self, value: pyquil.Program) -> None:
        if not isinstance(value, pyquil.Program):
            raise ValueError("Program must be an instance of pyquil.Program")
        self._program = value

    @property
    def qubits(self) -> List[int]:
        """Return the qubits acted upon by the operations in this circuit"""
        return list(self.program.get_qubits())

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits in the circuit."""
        return len(self.qubits)

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return 0

    @property
    def depth(self) -> int:
        """Return the circuit depth (i.e., length of critical path)."""
        return len(self.program)

    def _unitary(self) -> "np.ndarray":
        """Return the unitary of a pyQuil program."""
        return program_unitary(self.program, n_qubits=self.num_qubits)

    def remove_idle_qubits(self) -> None:
        """Checks whether the circuit uses contiguous qubits/indices,
        and if not, reduces dimension accordingly."""
        raise NotImplementedError

    def reverse_qubit_order(self) -> None:
        """Reverse the order of the qubits in the circuit."""
        raise NotImplementedError



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing fixtures for testing

"""
import pytest

from .braket.circuits import braket_bell, braket_shared15
from .cirq.circuits import cirq_bell, cirq_shared15
from .pennylane.circuits import pennylane_bell, pennylane_shared15
from .pyquil.circuits import pyquil_bell, pyquil_shared15
from .pytket.circuits import pytket_bell, pytket_shared15
from .qasm2.circuits import qasm2_bell, qasm2_cirq_shared15
from .qasm3.circuits import qasm3_bell, qasm3_shared15
from .qiskit.circuits import qiskit_bell, qiskit_shared15

# Map package names to their bell circuit functions
bell_circuit_functions = {
    "braket": braket_bell,
    "pennylane": pennylane_bell,
    "cirq": cirq_bell,
    "pyquil": pyquil_bell,
    "qiskit": qiskit_bell,
    "pytket": pytket_bell,
    "qasm2": qasm2_bell,
    "qasm3": qasm3_bell,
}

# Map package names to their shared15 circuit functions
shared15_circuit_functions = {
    "braket": braket_shared15,
    "pennylane": pennylane_shared15,
    "cirq": cirq_shared15,
    "pyquil": pyquil_shared15,
    "qiskit": qiskit_shared15,
    "pytket": pytket_shared15,
    "qasm2": qasm2_cirq_shared15,
    "qasm3": qasm3_shared15,
}


@pytest.fixture
def bell_circuit(request):
    """Fixture for getting bell circuit for a given package."""
    package = request.param
    circuit_func = bell_circuit_functions[package]
    circuit = circuit_func()
    return circuit, package


@pytest.fixture
def two_bell_circuits(request):
    """Fixture for getting two bell circuits from two different packages."""
    package1, package2 = request.param
    circuit1 = bell_circuit_functions[package1]()
    circuit2 = bell_circuit_functions[package2]()
    return circuit1, circuit2, package1, package2


@pytest.fixture
def bell_unitary():
    """Fixture containing unitary for bell circuit."""
    circuit = cirq_bell()
    return circuit.unitary()


@pytest.fixture
def shared15_circuit(request):
    """Fixture for getting shared15 circuit and unitary for a specific package."""
    package = request.param
    circuit_func = shared15_circuit_functions[package]
    circuit = circuit_func()
    return circuit, package


@pytest.fixture
def two_shared15_circuits(request):
    """Fixture for getting two shared15 circuits from two different packages."""
    package1, package2 = request.param
    circuit1 = shared15_circuit_functions[package1]()
    circuit2 = shared15_circuit_functions[package2]()
    return circuit1, circuit2, package1, package2


@pytest.fixture
def shared15_unitary():
    """Fixture containing unitary for shared15 circuit."""
    circuit = cirq_shared15()
    return circuit.unitary()


packages_bell = list(bell_circuit_functions.keys())
packages_shared15 = list(shared15_circuit_functions.keys())



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining abstract QuantumJobResult Class

"""
from abc import ABC, abstractmethod


class QuantumJobResult(ABC):
    """Abstract interface for result-like classes.

    Args:
        _result: A result-like object

    """

    def __init__(self, _result):
        self._result = _result

    @abstractmethod
    def measurements(self):
        """Return measurements as list"""

    @abstractmethod
    def raw_counts(self):
        """Returns raw histogram data of the run"""

    @staticmethod
    def format_counts(counts: dict, include_zero_values: bool = False) -> dict:
        """Formats, sorts, and adds missing bit indices to counts dictionary
        Can pass in a 'include_zero_values' parameter to decide whether to include the states
        with zero counts.

        For example:

        .. code-block:: python

            >>> counts
            {'1 1': 13, '0 0': 46, '1 0': 79}
            >>> QuantumJobResult.format_counts(counts)
            {'00': 46, '10': 79, '11': 13}
            >>> QuantumJobResult.format_counts(counts, include_zero_values=True)
            {'00': 46, '01': 0, '10': 79, '11': 13}

        """
        counts = {key.replace(" ", ""): value for key, value in counts.items()}

        num_bits = max(len(key) for key in counts)
        all_keys = [format(i, f"0{num_bits}b") for i in range(2**num_bits)]
        final_counts = {key: counts.get(key, 0) for key in sorted(all_keys)}

        if not include_zero_values:
            final_counts = {key: value for key, value in final_counts.items() if value != 0}

        return final_counts

    def measurement_counts(self, include_zero_values: bool = False) -> dict:
        """Returns the sorted histogram data of the run"""
        raw_counts = self.raw_counts()
        if isinstance(raw_counts, dict):
            return self.format_counts(raw_counts, include_zero_values=include_zero_values)
        return [
            self.format_counts(counts, include_zero_values=include_zero_values)
            for counts in raw_counts
        ]



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint:disable=invalid-name

"""
Module defining abstract QuantumDevice Class

"""

import json
import warnings
from abc import ABC, abstractmethod
from datetime import datetime
from typing import TYPE_CHECKING, Dict  # pylint: disable=unused-import

from qbraid.api import ApiError, QbraidSession
from qbraid.exceptions import QbraidError
from qbraid.load_program import circuit_wrapper
from qbraid.transpiler.exceptions import CircuitConversionError

from .enums import JobStatus
from .exceptions import ProgramValidationError, QbraidRuntimeError

if TYPE_CHECKING:
    import qbraid


class QuantumDevice(ABC):
    """Abstract interface for device-like classes."""

    def __init__(self, device: "qbraid.QDEVICE"):
        """Create a ``QuantumDevice`` object.

        Args:
            device (:data:`~.qbraid.QDEVICE`): qBraid Quantum device object

        """
        # pylint: disable=too-many-function-args
        self._device = device
        self._id = None
        self._name = None
        self._vendor = None
        self._provider = None
        self._num_qubits = None
        self._device_type = None
        self._run_package = None
        self._populate_metadata(device)

    @property
    def id(self) -> str:
        """Return the device ID."""
        return self._id

    @property
    def name(self) -> str:
        """Return the device name.

        Returns:
            The name of the device.
        """
        return self._name

    @property
    def provider(self) -> str:
        """Return the device provider.

        Returns:
            The provider responsible for the device.

        """
        return self._provider

    @property
    def vendor(self) -> str:
        """Return the software vendor name.

        Returns:
            The name of the software vendor.

        """
        return self._vendor

    @property
    def num_qubits(self) -> int:
        """The number of qubits supported by the device.

        Returns:
            Number of qubits supported by QPU. If Simulator returns None.

        """
        return self._num_qubits

    @property
    def device_type(self) -> "qbraid.providers.DeviceType":
        """The device type, Simulator, Fake_device or QPU.

        Returns:
            Device type enum (SIMULATOR|QPU|FAKE)

        """
        return self._device_type

    @abstractmethod
    def status(self) -> "qbraid.providers.DeviceStatus":
        """Return device status."""

    @abstractmethod
    def queue_depth(self) -> int:
        """Return the number of jobs in the queue for the backend"""

    @abstractmethod
    def _populate_metadata(self, device: "qbraid.QDEVICE") -> None:
        """Populate device metadata with the following fields:
        * self._id
        * self._name
        * self._provider
        * self._vendor
        * self._num_qubits
        * self._device_type
        """

    def metadata(self) -> dict:
        """Returns device metadata"""
        return {
            "id": self._id,
            "name": self._name,
            "provider": self._provider,
            "vendor": self._vendor,
            "numQubits": self._num_qubits,
            "deviceType": self._device_type.name,
            "status": self.status().name,
            "queueDepth": self.queue_depth(),
        }

    def __str__(self):
        return f"{self.vendor} {self.provider} {self.name} device wrapper"

    def __repr__(self):
        """String representation of a DeviceWrapper object."""
        return f"<{self.__class__.__name__}({self.provider}:'{self.name}')>"

    def verify_run(self, run_input: "qbraid.QPROGRAM") -> None:
        """Checks device is online and that circuit num qubits <= device num qubits.

        Raises:
            QbraidRuntimeError: If error applying circuit wrapper or circuit number of
                qubits exceeds device number qubits

        """
        if self.status().value == 1:
            warnings.warn(
                "Device is currently offline. Depending on the provider queueing system, "
                "submitting this job may result in an exception or a long wait time.",
                UserWarning,
            )

        try:
            qbraid_circuit = circuit_wrapper(run_input)
        except QbraidError as err:
            raise ProgramValidationError from err

        if self.num_qubits and qbraid_circuit.num_qubits > self.num_qubits:
            raise ProgramValidationError(
                f"Number of qubits in circuit ({qbraid_circuit.num_qubits}) exceeds "
                f"number of qubits in device ({self.num_qubits})."
            )

    def transpile(self, run_input: "qbraid.QPROGRAM") -> "qbraid.QPROGRAM":
        """Convert circuit to package compatible with target device and pass through any
        provider transpile methods to match topology of device and/or optimize as applicable.

        Returns:
            :data:`~qbraid.QPROGRAM`: Transpiled quantum program (circuit) object

        Raises:
            QbraidRuntimeError: If circuit conversion fails

        """
        input_run_package = run_input.__module__.split(".")[0]
        if input_run_package != self._run_package:
            qbraid_circuit = circuit_wrapper(run_input)
            try:
                run_input = qbraid_circuit.transpile(self._run_package)
            except CircuitConversionError as err:
                raise QbraidRuntimeError from err
        return self._transpile(run_input)

    def compile(self, run_input: "qbraid.QPROGRAM") -> "qbraid.QPROGRAM":
        """Compile run input.

        Returns:
            :data:`~qbraid.QPROGRAM`: Compiled quantum program (circuit) object

        Raises:
            QbraidRuntimeError: If circuit conversion fails

        """
        return self._compile(run_input)

    def process_run_input(
        self, run_input: "qbraid.QPROGRAM", auto_compile: bool = False
    ) -> "qbraid.transpiler.QuantumProgram":
        """Process quantum program before passing to device run method.

        Returns:
            :class:`~qbraid.transpiler.QuantumProgram`: qBraid wrapped quantum program object

        Raises:
            QbraidRuntimeError: If error processing run input

        """
        self.verify_run(run_input)
        run_input = self.transpile(run_input)
        if auto_compile:
            run_input = self._compile(run_input)
        return circuit_wrapper(run_input)

    def _init_job(
        self,
        vendor_job_id: str,
        circuits: "qbraid.transpiler.QuantumProgram",
        shots: int,
        tags: Dict[str, str],
    ) -> str:
        """Initialize data dictionary for new qbraid job and
        create associated MongoDB job document.

        Args:
            vendor_job_id: Job ID provided by device vendor
            circuit: Wrapped quantum circuit list
            shots: Number of shots

        Returns:
            The qbraid job ID associated with this job

        """
        session = QbraidSession()

        vendor = self.vendor.lower()
        # One of the features of qBraid Quantum Jobs is the ability to send
        # jobs without any credentials using the qBraid Lab platform. If the
        # qBraid Quantum Jobs proxy is enabled, a document has already been
        # created for this job. So, instead creating a duplicate, we query the
        # user jobs for the `vendorJobId` and return the correspondong `qbraidJobId`.
        if session._running_in_lab() and session._qbraid_jobs_enabled(vendor):
            try:
                job = session.post("/get-user-jobs", json={"vendorJobId": vendor_job_id}).json()[0]
                return job.get("qbraidJobId", job.get("_id"))
            except IndexError as err:
                raise ApiError(f"{self.vendor} job {vendor_job_id} not found") from err

        # get qBraid device ID. Temporary workaround until we have a better way
        qbraid_id = session.get("/public/lab/get-devices", params={"objArg": self.id}).json()[0][
            "qbraid_id"
        ]

        # Create a new document for the user job. The qBraid API creates a unique
        # Job ID, which is collected in the response. We use dummy variables for
        # each of the status fields, which will be updated via the `get_job_data`
        # function upon instantiation of the `QuantumJob` object.
        init_data = {
            "qbraidJobId": "",
            "vendorJobId": vendor_job_id,
            "qbraidDeviceId": qbraid_id,
            "vendorDeviceId": self.id,
            "shots": shots,
            "tags": json.dumps(tags),
            "createdAt": datetime.utcnow(),
            "status": "UNKNOWN",  # this will be set after we get back the job ID and check status
            "qbraidStatus": JobStatus.INITIALIZING.name,
            "email": session.user_email,
        }

        if len(circuits) == 1:
            init_data["circuitNumQubits"] = circuits[0].num_qubits
            init_data["circuitDepth"]: circuits[0].depth
        else:
            init_data["circuitBatchNumQubits"] = ([circuit.num_qubits for circuit in circuits],)
            init_data["circuitBatchDepth"] = [circuit.depth for circuit in circuits]

        return session.post("/init-job", data=init_data).json()

    @abstractmethod
    def _transpile(self, run_input):
        """Applies any software/device specific modifications to run input."""

    @abstractmethod
    def _compile(self, run_input):
        """Applies any software/device specific modifications to run input."""

    @abstractmethod
    def run(self, run_input: "qbraid.QPROGRAM", *args, **kwargs) -> "qbraid.providers.QuantumJob":
        """Abstract run method."""



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing OpenQASM conversion function

"""
import os

import openqasm3

from qbraid.inspector import get_qasm_version
from qbraid.transpiler.conversions.qasm_qelib1 import _decompose_rxx_instr

QASMType = str


def _get_qasm3_gate_defs() -> str:
    """Helper function to get openqasm 3 gate defs from .qasm file"""
    current_dir = os.path.dirname(os.path.abspath(__file__))
    qelib_qasm = os.path.join(current_dir, "qelib_qasm3.qasm")
    with open(qelib_qasm, mode="r", encoding="utf-8") as file:
        gate_defs = file.read()
    return gate_defs


def _build_qasm_3_reg(line: str, qreg_type: bool) -> QASMType:
    """Helper function to build openqasm 3 register statements

    Args:
        line (str): openqasm 2  regdecl statement
        qreg_type (bool): whether a qreg or creg type statement

    Returns:
        str : openqasm 3 qubits / bits declaration
    """
    reg_keyword_len = 4
    line = line[reg_keyword_len:]
    elements = line.split("[")
    reg_name = elements[0].strip()
    reg_size = elements[1].split("]")[0].strip()
    result = "qubit" if qreg_type else "bit"
    result += f"[{reg_size}] {reg_name};\n"
    return result


def _build_qasm_3_measure(line: str) -> QASMType:
    """Helper function to build openqasm 3 measure string

    Args:
        line (str): openqasm 2 measure statement

    Returns:
        str:  openqasm 3 measure statement
    """
    measure_keyword_len = 7
    line = line[measure_keyword_len:]
    elements = line.split("->")
    qubits_name = elements[0].replace(" ", "")
    bits_name = elements[1].split(";")[0].replace(" ", "")

    return f"{bits_name} = measure {qubits_name};\n"


def _convert_line_to_qasm3(line: str) -> QASMType:
    """Function to change an openqasm 2 line to openqasm 3

    Args:
        line (str): an openqasm 2 line

    Returns:
        str: corresponding openqasm 3 line
    """
    line = line.lstrip()
    if line.startswith("OPENQASM"):
        return ""
    if "qelib1.inc" in line:
        return ""
    if line.startswith("qreg"):
        return _build_qasm_3_reg(line, qreg_type=True)
    if line.startswith("creg"):
        return _build_qasm_3_reg(line, qreg_type=False)
    if line.startswith("u("):
        return line.replace("u(", "U(")
    if line.startswith("rxx("):
        return _decompose_rxx_instr(line)
    if line.startswith("measure"):
        return _build_qasm_3_measure(line)
    if line.startswith("opaque"):
        # as opaque is ignored by openqasm 3 add it as a comment
        return "// " + line + "\n"
    return line + "\n"


def qasm2_to_qasm3(qasm_str: str) -> QASMType:
    """Convert a OpenQASM 2.0 string to OpenQASM 3.0 string

    Args:
        qasm_str (str): OpenQASM 2.0 string

    Returns:
        str: OpenQASM 3.0 string
    """
    qasm_version = get_qasm_version(qasm_str)
    if not qasm_version == "qasm2":
        raise ValueError("Invalid OpenQASM 2.0 string")

    qasm3_str = "OPENQASM 3.0;\ninclude 'stdgates.inc';"

    gate_defs = _get_qasm3_gate_defs()

    for line in gate_defs:
        qasm3_str += line

    for line in qasm_str.splitlines():
        line = _convert_line_to_qasm3(line)
        qasm3_str += line

    return qasm3_str


def qasm3_to_openqasm3(qasm_str: str) -> openqasm3.ast.Program:
    """Loads an openqasm3.ast.Program from an OpenQASM 3.0 string

    Args:
        qasm_str (str): OpenQASM 3.0 string

    Returns:
        openqasm3.ast.Program: OpenQASM 3.0 AST program
    """
    return openqasm3.parse(qasm_str)


def openqasm3_to_qasm3(program: openqasm3.ast.Program) -> str:
    """Dumps openqasm3.ast.Program to an OpenQASM 3.0 string

    Args:
        program (openqasm3.ast.Program): OpenQASM 3.0 AST program

    Returns:
        str: OpenQASM 3.0 string
    """
    return openqasm3.dumps(program)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
PyTKET conversions

.. currentmodule:: qbraid.transpiler.conversions.pytket

Functions
----------

.. autosummary::
   :toctree: ../stubs/

   pytket_to_qasm2
   qasm2_to_pytket

"""
from qbraid.transpiler.conversions.pytket.conversions_qasm import pytket_to_qasm2, qasm2_to_pytket



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining BraketQuantumTask Class

"""
import logging
from typing import Optional

from braket.aws import AwsQuantumTask

from qbraid.providers.enums import JOB_FINAL
from qbraid.providers.exceptions import JobStateError
from qbraid.providers.job import QuantumJob

from .result import BraketGateModelResult
from .tracker import get_quantum_task_cost


class AmazonBraketVersionError(Exception):
    """Exception raised for Amazon Braket SDK errors due to versioning."""


class BraketQuantumTask(QuantumJob):
    """Wrapper class for Amazon Braket ``QuantumTask`` objects."""

    def __init__(self, job_id: str, **kwargs):
        """Create a BraketQuantumTask."""

        super().__init__(job_id, **kwargs)

    def _get_job(self):
        """Return the job like object that is being wrapped."""
        return AwsQuantumTask(self.vendor_job_id)

    def _get_status(self):
        """Returns status from Braket QuantumTask object metadata."""
        return self._job.state()

    def queue_position(self) -> Optional[int]:
        """Returns queue position from Braket QuantumTask.
        '>2000' returns as 2000 for typing consistency."""
        try:
            position = self._job.queue_position().queue_position
            if isinstance(position, str):
                if position.startswith(">"):
                    position = position[1:]
                return int(position)
            return position
        except AttributeError as err:
            raise AmazonBraketVersionError(
                "Queue visibility is only available for amazon-braket-sdk>=1.56.0"
            ) from err

    def result(self) -> BraketGateModelResult:
        """Return the results of the job."""
        if self.status() not in JOB_FINAL:
            logging.info("Result will be available when job has reached final state.")
        return BraketGateModelResult(self._job.result())

    def cancel(self) -> None:
        """Cancel the quantum task."""
        task = self._job
        status = self.status()
        if status in JOB_FINAL:
            raise JobStateError(f"Cannot cancel quantum job in the {status} state.")
        try:
            task.cancel()
        except RuntimeError:
            task._aws_session.cancel_quantum_task(task.arn)

    def get_cost(self) -> float:
        """Return the cost of the job."""
        decimal_cost = get_quantum_task_cost(self.vendor_job_id)
        return float(decimal_cost)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# isort: skip_file
# pylint: skip-file

"""
Mdule submiting and managing jobs through IBM and IBM backends.

.. currentmodule:: qbraid.providers.ibm

Classes
--------

.. autosummary::
   :toctree: ../stubs/

   QiskitBackend
   QiskitJob
   QiskitRemoteService
   QiskitProvider
   QiskitRuntime
   QiskitResult

"""
from .result import QiskitResult
from .device import QiskitBackend
from .job import QiskitJob
from .provider import QiskitRemoteService, QiskitProvider, QiskitRuntime



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for configuring provider credentials and authentication.

"""

import os
from typing import TYPE_CHECKING, List, Optional

import boto3
from boto3.session import Session
from braket.aws import AwsDevice, AwsSession

from qbraid.api import QbraidSession
from qbraid.exceptions import QbraidError

if TYPE_CHECKING:
    import braket.aws


class BraketProvider:
    """
    This class is responsible for managing the interactions and
    authentications with the AWS services.

    Attributes:
        aws_access_key_id (str): AWS access key ID for authenticating with AWS services.
        aws_secret_access_key (str): AWS secret access key for authenticating with AWS services.
    """

    REGIONS = ("us-east-1", "us-west-1", "us-west-2", "eu-west-2")

    def __init__(self, aws_access_key_id=None, aws_secret_access_key=None):
        """
        Initializes the QbraidProvider object with optional AWS credentials.

        Args:
            aws_access_key_id (str, optional): AWS access key ID. Defaults to None.
            aws_secret_access_key (str, optional): AWS secret access token. Defaults to None.
        """
        self.aws_access_key_id = aws_access_key_id or os.getenv("AWS_ACCESS_KEY_ID")
        self.aws_secret_access_key = aws_secret_access_key or os.getenv("AWS_SECRET_ACCESS_KEY")

    def save_config(self):
        """Save the current configuration."""
        raise NotImplementedError

    @staticmethod
    def _get_default_region() -> str:
        """Returns the default AWS region."""
        try:
            session = Session()
            return session.region_name
        except Exception:  # pylint: disable=broad-exception-caught
            return "us-east-1"

    @staticmethod
    def _get_s3_default_bucket() -> str:
        """Return name of the default bucket to use in the AWS Session"""
        try:
            aws_session = AwsSession()
            return aws_session.default_bucket()
        except Exception:  # pylint: disable=broad-exception-caught
            return "amazon-braket-qbraid-provider"

    def _get_region_name(self, device_arn: str) -> str:
        """Returns the AWS region name."""
        maybe_region = device_arn.split(":")[3]
        if maybe_region in self.REGIONS:
            return maybe_region

        provider = device_arn.split("/")[-2]
        if provider in ["ionq", "quera", "xanadu"]:
            return "us-east-1"
        if provider == "oqc":
            return "eu-west-2"
        if provider == "rigetti":
            return "us-west-1"

        return self._get_default_region()

    def _get_aws_session(self, region_name: Optional[str] = None) -> "braket.aws.AwsSession":
        """Returns the AwsSession provider."""
        region_name = region_name or self._get_default_region()
        default_bucket = self._get_s3_default_bucket()

        boto_session = Session(
            region_name=region_name,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
        )
        return AwsSession(boto_session=boto_session, default_bucket=default_bucket)

    def get_devices(
        self, aws_session=None, statuses=None, **kwargs
    ) -> List["braket.aws.AwsDevice"]:
        """Return a list of backends matching the specified filtering."""
        aws_session = self._get_aws_session() if aws_session is None else aws_session
        statuses = ["ONLINE", "OFFLINE"] if statuses is None else statuses

        return AwsDevice.get_devices(aws_session=aws_session, statuses=statuses, **kwargs)

    def get_device(self, device_arn: str) -> "braket.aws.AwsDevice":
        """Returns the AWS device."""
        region_name = self._get_region_name(device_arn)
        aws_session = self._get_aws_session(region_name=region_name)
        return AwsDevice(arn=device_arn, aws_session=aws_session)

    def get_tasks_by_tag(
        self, key: str, values: Optional[List[str]] = None, region_names: Optional[List[str]] = None
    ) -> List[str]:
        """
        Retrieves a list of quantum task ARNs that match the specified tag keys or key/value pairs.

        Args:
            key (str): The tag key to match.
            values (Optional[List[str]]): A list of tag values to match against the provided
                                          key. If None, tasks with the specified key,
                                          regardless of its value, are matched.
            region_names (Optional[List[str]]): A list of region names to search. If None, all
                                                regions in `self.REGIONS` are searched.

        Returns:
            List[str]: A list of ARNs for quantum tasks that match the given tag criteria.

        Raises:
            QbraidError: If the function is called within a qBraid quantum job environment
                         where AWS S3 requests are not supported.
        """
        if QbraidSession._running_in_lab() and QbraidSession._qbraid_jobs_enabled():
            raise QbraidError("AWS S3 requests not supported by qBraid quantum jobs.")

        region_names = (
            region_names
            if region_names is not None and len(region_names) > 0
            else list(self.REGIONS)
        )
        values = values if values is not None else []

        tasks = []
        for region_name in region_names:
            client = boto3.client("resourcegroupstaggingapi", region_name=region_name)
            response = client.get_resources(
                TagFilters=[
                    {
                        "Key": key,
                        "Values": values,
                    }
                ],
            )
            matches = [t["ResourceARN"] for t in response["ResourceTagMappingList"]]
            tasks += matches

        return tasks



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module that implements qelib1.inc qasm gate definitions as python functions

"""
import re
from typing import Optional

from qbraid.exceptions import QasmError


def _get_param(instr: str) -> Optional[str]:
    try:
        return instr[instr.index("(") + 1 : instr.index(")")]
    except ValueError:
        return None


def _remove_spaces_in_parentheses(expression):
    # Find all parenthesized parts of the input.
    parenthesized_parts = re.findall(r"\(.*?\)", expression)

    for part in parenthesized_parts:
        # For each parenthesized part, remove all the spaces.
        cleaned_part = part.replace(" ", "")

        # Replace the original part in the expression with the cleaned part.
        expression = expression.replace(part, cleaned_part)

    return expression


def _decompose_cu_instr(instr: str) -> str:
    """controlled-U gate"""
    try:
        instr = _remove_spaces_in_parentheses(instr)
        cu_gate, qs = instr.split(" ")
        a, b = qs.strip(";").split(",")
        params_lst = _get_param(cu_gate).split(",")
        params = [float(x) for x in params_lst]
        theta, phi, lam, gamma = params
    except (AttributeError, ValueError) as err:
        raise QasmError from err
    instr_out = "\n// cu gate\n"
    instr_out += f"p({gamma}) {a};\n"
    instr_out += f"p({(lam+phi)/2}) {a};\n"
    instr_out += f"p({(lam-phi)/2}) {b};\n"
    instr_out += f"cx {a},{b};\n"
    instr_out += f"u({-1*theta/2},0,{-1*(phi+lam)/2}) {b};\n"
    instr_out += f"cx {a},{b};\n"
    instr_out += f"u({theta/2},{phi},0) {b};\n\n"
    return instr_out


def _decompose_rxx_instr(instr: str) -> str:
    """two-qubit XX rotation"""
    try:
        instr = instr.replace(", ", ",")
        rxx_gate, qs = instr.split(" ")
        a, b = qs.strip(";").split(",")
        theta = _get_param(rxx_gate)
    except (AttributeError, ValueError) as err:
        raise QasmError from err
    instr_out = "\n// rxx gate\n"
    instr_out += f"h {a};\n"
    instr_out += f"h {b};\n"
    instr_out += f"cx {a},{b};\n"
    instr_out += f"rz({theta}) {b};\n"
    instr_out += f"cx {a},{b};\n"
    instr_out += f"h {b};\n"
    instr_out += f"h {a};\n\n"
    return instr_out


def _decompose_rccx_instr(instr: str) -> str:
    """relative-phase CCX"""
    try:
        _, qs = instr.split(" ")
        a, b, c = qs.strip(";").split(",")
    except (AttributeError, ValueError) as err:
        raise QasmError from err
    instr_out = "\n// rccx gate\n"
    instr_out += f"u2(0,pi) {c};\n"
    instr_out += f"u1(pi/4) {c};\n"
    instr_out += f"cx {b},{c};\n"
    instr_out += f"u1(-pi/4) {c};\n"
    instr_out += f"cx {a},{c};\n"
    instr_out += f"u1(pi/4) {c};\n"
    instr_out += f"cx {b},{c};\n"
    instr_out += f"u1(-pi/4) {c};\n"
    instr_out += f"u2(0,pi) {c};\n\n"
    return instr_out


def _decompose_rc3x_instr(instr: str) -> str:
    """relative-phase 3-controlled X gate"""
    try:
        _, qs = instr.split(" ")
        a, b, c, d = qs.strip(";").split(",")
    except (AttributeError, ValueError) as err:
        raise QasmError from err
    instr_out = "\n// rc3x gate\n"
    instr_out += f"u2(0,pi) {d};\n"
    instr_out += f"u1(pi/4) {d};\n"
    instr_out += f"cx {c},{d};\n"
    instr_out += f"u1(-pi/4) {d};\n"
    instr_out += f"u2(0,pi) {d};\n"
    instr_out += f"cx {a},{d};\n"
    instr_out += f"u1(pi/4) {d};\n"
    instr_out += f"cx {b},{d};\n"
    instr_out += f"u1(-pi/4) {d};\n"
    instr_out += f"cx {a},{d};\n"
    instr_out += f"u1(pi/4) {d};\n"
    instr_out += f"cx {b},{d};\n"
    instr_out += f"u1(-pi/4) {d};\n"
    instr_out += f"u2(0,pi) {d};\n"
    instr_out += f"u1(pi/4) {d};\n"
    instr_out += f"cx {c},{d};\n"
    instr_out += f"u1(-pi/4) {d};\n"
    instr_out += f"u2(0,pi) {d};\n\n"
    return instr_out


def decompose_qasm_qelib1(qasm_str: str) -> str:
    """Replace edge-case qelib1 gates with equivalent decomposition."""
    qasm_lst_out = []
    qasm_lst = qasm_str.split("\n")

    for _, qasm_line in enumerate(qasm_lst):
        line_str = qasm_line
        len_line = len(line_str)
        if len_line > 3 and line_str[0:3] == "cu(":
            line_str_out = _decompose_cu_instr(line_str)
        elif len_line > 4 and line_str[0:4] == "rxx(":
            line_str_out = _decompose_rxx_instr(line_str)
        elif len_line > 4 and line_str[0:4] == "rccx":
            line_str_out = _decompose_rccx_instr(line_str)
        elif len_line > 4 and line_str[0:4] == "rc3x":
            line_str_out = _decompose_rc3x_instr(line_str)
        else:
            line_str_out = line_str

        qasm_lst_out.append(line_str_out)

    qasm_str_def = "\n".join(qasm_lst_out)
    return qasm_str_def
